<%
  ' clsReport.asp
  
  Class clsReport
    Dim columns()
    Dim column_count
    Dim groups()
    Dim group_count
    Dim column_divider, row_start, row_end, group_row_start, group_row_end, group_divider, blank
    Dim header_start, header_end
    
    Private Sub Class_Initialize
      ReDim columns(0)
      column_count = 0
      ReDim groups(0)
      group_count = 0
      
      column_divider = vbTab
      row_start = ""
      row_end = vbCRLF
      group_row_start = ""
      group_row_end = vbCRLF
      header_start = ""
      header_end = vbCRLF
      blank = ""
    End Sub
    
    Public Function AddColumn(name, title, format)
      Dim c
      Set c = new clsReportColumn
      c.fieldname = name
      c.fieldtitle = title
      c.format.datatype = format
      
      ReDim Preserve columns(column_count)
      Set columns(column_count) = c
      
      column_count = column_count + 1
      Set AddColumn = c
    End Function

    Public Function AddGroup(columnname, groupname)
      Dim g
      Set g = new clsReportGroup
      g.column = columnname
      
      ReDim Preserve groups(group_count)
      Set groups(group_count) = g
      
      If group_count > 0 Then
        Set g.prev_group = groups(group_count - 1)
      End If
      
      g.name = groupname
      
      group_count = group_count + 1
      Set AddGroup = g
    End Function
    
    Public Function Generate(ByRef oRS)
      UpdateAggregateColumnNumbers
      Generate = GenerateColumnHeaders & GenerateGroup(oRS, 0)
    End Function
    
    Private Function GenerateGroup(ByRef oRS, level)
      Dim str, first, i, done, value
      str = ""
      ' We only need to track the group starts and stops if we are not
      ' at the root level.
      If level <> 0 Then
        first_value = oRS(groups(level - 1).column) & ""
        groups(level - 1).ResetAggregates
      End If
      done = False
      Do While NOT oRS.EOF AND NOT done
        ' While we're not out of records, advance forward
        If level < group_count Then
          ' There's another level deeper, so let's recurse into it!
          str = str & GenerateGroup(oRS, level + 1)
        Else
          ' We're at the lowest level... let's add some output
          str = str & row_start
          For i = 0 to column_count - 1
            If columns(i).display Then
              value = Trim(columns(i).ToString(oRS))
              If value = "" Then
                str = str & blank
              Else 
                str = str & value
              End If
              If i < column_count - 1 Then
                str = str & column_divider
              End If
            End If
          Next
          groups(level - 1).UpdateAggregates oRS
          str = str & row_end
          oRS.MoveNext
        End If
        If NOT oRS.EOF AND level > 0 Then
          done = (oRS(groups(level - 1).column) & "") <> first_value
        End If
      Loop
      
      ' Add the group totals
      
      If level > 0 Then
        If NOT groups(level - 1).display Then
          str = ""
        End If
        str = str & GenerateGroupTotals(level - 1)
      End If
      
      GenerateGroup = str
    End Function
    
    Private Function GenerateColumnHeaders()
      Dim str
      str = header_start
      For i = 0 to column_count - 1
        If columns(i).display Then
          str = str & columns(i).fieldtitle
          If i < column_count - 1 Then
            str = str & column_divider
          End If
        End If
      Next
      str = str & header_end
      GenerateColumnHeaders = str
    End Function
    
    Private Function GenerateGroupTotals(level)
      Dim str, arr(), i, c_number, value, name
      ' Create an array big enough to hold all the columns
      ReDim arr(column_count)
      name = groups(level).name
      
      For i = 0 to column_count - 1
        arr(i) = blank
      Next
      
      arr(groups(level).name_column_number) = name
      ' Loop through all the aggregates in the group and put popuplate the array
      For i = 0 To groups(level).aggregate_count - 1
        c_number = groups(level).aggregates(i).column_number
        value = groups(level).aggregates(i).total
        ' Use the column to format the aggregate's value
        value = columns(c_number).format.FormatString(value)
        ' Put the value back into the array
        arr(c_number) = value
      Next
      
      tstr = ""
      For i = 0 to column_count - 1
        If columns(i).display Then
          If tstr <> "" Then
            tstr = tstr & group_divider
          End If
          tstr = tstr & arr(i)
        End If
      Next
      str = group_row_start & tstr & group_row_end
      GenerateGroupTotals = str
    End Function
    
    Private Function UpdateAggregateColumnNumbers()
      ' This function updates the aggregates so that they know which column in the report
      ' they are based upon the columns added to the report.
      Dim hash
      Dim i, j, col
      
      Set hash = Server.CreateObject("Scripting.Dictionary")
      
      For i = 0 to column_count - 1
        hash.Add columns(i).fieldname, i
      Next
      
      For i = 0 to group_count - 1
        For j = 0 to groups(i).aggregate_count - 1
          groups(i).aggregates(j).column_number = hash.Item(groups(i).aggregates(j).fieldname)
        Next
      Next
    End Function
    
  End Class
  
  Class clsReportColumn
    Public fieldname, fieldtitle
    Public format, display
    
    Private Sub Class_Initialize
      fieldformat = "string"
      Set format = New clsReportColumnFormat
      display = True
    End Sub
    
    Public Function ToString(ByRef oRS)
      value = "" & oRS(fieldname)
      ToString = format.FormatString(value)
    End Function
    
  End Class
  
  Class clsReportColumnFormat
    Dim template, datatype
    
    Private Sub Class_Initialize
      template = "[value]"
    End Sub
    
    Public Function FormatString(value)
      Dim str
      If datatype = "string" Then
        str = value
      ElseIf datatype = "numberformat" Then
        str = FormatNumber(value, 0)
      ElseIf datatype = "number" Then
        str = CDbl(value)
      ElseIf datatype = "money" Then
        str = FormatCurrency(value)
      ElseIf datatype = "shortdate" Then
        str = FormatDateTime(value, vbShortDate)
      End If
      str = Replace(template, "[value]", str)
      FormatString = str
    End Function
  End Class
  
  Class clsReportGroup
    Public column, name, display
    Public aggregates()
    Public aggregate_count
    Public prev_group
    Public name_column_number
    
    Private Sub Class_Initialize
      ReDim aggregates(0)
      aggregate_count = 0
      Set prev_group = Nothing
      name_column_number = 0
      display = True
    End Sub

    Public Function AddAggregate(fieldname, method)
      Dim a
      Set a = new clsReportAggregate
      a.fieldname = fieldname
      a.method = method
      
      ReDim Preserve aggregates(aggregate_count)
      Set aggregates(aggregate_count) = a
      
      aggregate_count = aggregate_count + 1
      Set AddAggregate = a
    End Function
    
    Public Function UpdateAggregates(ByRef oRS)
      Dim i
      
      ' Update this gruops aggregates
      For i = 0 To aggregate_count - 1
        aggregates(i).Update self, oRS
      Next
      
      ' Update all parent aggregrates... this will work recursively
      Set t = prev_group
      If NOT t IS Nothing Then
        t.UpdateAggregates oRS
      End If
    End Function

    Public Function ResetAggregates()
      Dim i
      For i = 0 To aggregate_count - 1
        aggregates(i).Reset
      Next
    End Function
  End Class
  
  Class clsReportAggregate
    Public fieldname, method
    Public total, count
    Public column_number
    
    Private Sub Class_Initialize
      method = "sum"
      column_number = 0
      Reset()
    End Sub
    
    Public Sub Update(ByRef oGroup, ByRef oRS)
      If method = "first" Then
        total = oRS(fieldname)
      End If
      If method = "sum" Then
        total = total + CDbl(oRS(fieldname))
      End If
      If method = "avg" Then
        total = (total * count + CDbl(oRS(fieldname))) / (count + 1)
      End If
      If method = "count" Then
        total = total + 1
      End If
      count = count + 1
    End Sub
    
    Public Sub Reset()
      total = 0
      count = 0
    End Sub
  End Class
  
%>